#
# Module for functions interacting with the system.
#
############################################################
# Procedure to delete a file. Uses the builtin delete function if 
# the Tcl version is recent enough.
proc RemoveFile { name } {

    if { [info tclversion] > 7.5 } {
	file delete -force $name
    } else {
	exec /bin/rm -f $name
    }
}

############################################################
# Procedure to construct a file name from a name (usually the current
# basename) and an extension. If the initial name is empty, an empty name
# is returned
proc Make_File_Name { name extension } {

    set base [file rootname $name]

    if { [string length $base] > 0 } {
	return "$base.$extension"
    } else {
	return {}
    }
}

############################################################
# BrowseFiles -
# Procedure to create a file browser and let the user specify a
# files with the mouse. Used as a replacement for tk_getOpenFile or
# tk_getSaveFile when they are not available.
# Accordingly, a list of options is passed to the procedure. It is compatible
# with the one passed to the tk_getXxxxFile calls. However, only the 
# -defaultextension, -initialfile, -initialdir and -filetypes options are 
# recognized here.
#
# Parameters:
# - options	the list of options
#
proc BrowseFiles { options } {
    global BrowserAttributes 

    # Extensions will be extracted from the options, if they are there...
    set BrowserAttributes(DefaultExt) {}
    set BrowserAttributes(Types) {}
    
    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -defaultextension {
		incr i 1
		if { $i < [llength $options] } {
		    set opt [lindex $options $i]
		    set BrowserAttributes(DefaultExt) .$opt
		}
	    }
	    -initialfile {
		incr i 1
		if { $i < [llength $options] } {
		    set opt [lindex $options $i]
		    set BrowserAttributes(FileName) $opt
		}
	    }
	    -initialdir {
		incr i 1
		if { $i < [llength $options] } {
		    set opt [lindex $options $i]
		    set BrowserAttributes(Dir) $opt
		}
	    }
	    -filetypes {
		incr i 1
		if { $i < [llength $options] } {
		    set opt [lindex $options $i]
		    set BrowserAttributes(Types) $opt
		}
	    }
	}
    }

    # 
    # Make browser and let it grab focus
    set bname [MakeBrowser 2 $BrowserAttributes(Types)]
    grab $bname

    # Set initial dir listing then wait for the window to disappear
    set BrowserAttributes(Dir) [pwd]
    DisplayFiles $bname.list 2 \
	$BrowserAttributes(Dir) $BrowserAttributes(DefaultExt)
    tkwait window $bname 
    
    if [info exists BrowserAttributes(FileName)] {
	set file_name $BrowserAttributes(FileName)
    } else {
	set file_name ""
    }

    return $file_name
}

######################################################################
# DisplayFiles - 
# Procedure to display files in a set of "listbox-like" windows.
# Derived from the original procedure by Ch. Samtmann.
# Used in conjunction with "MakeBrowser" and "BrowseFiles"
# 
# Parameters:
# - win		an existing frame containing several text widgets
# - nbboxes	number of these boxes. The boxes are named $win.box0,
#		$win.box1 and so on.
# - dir		the directory listed
# - exts	list of lists of regular expressions used as filter.
#		The format is that of filetypes in tk_getxxxxFile.
#
proc DisplayFiles { win nbboxes dir exts }  {
    global BrowserAttributes
    
    # build a list of files according to the filter, but keeping 
    # directories, including "." and "..".
    if [catch {cd $dir} err] {
	puts "$err"
    }
    set BrowserAttributes(Dir) [pwd]
    set BrowserAttributes(DefaultExt) $exts
    set files [exec ls -a]
    set files_trim {}

    foreach i $files {
	if { [file isdirectory $i] } {
	    append i "/"
	    lappend files_trim $i
	} elseif { [llength $exts] > 0 } {
	    set good 0
	    foreach ext $exts {
		set good [expr $good \
			      || [string match $ext [file extension $i] ]]
	    }
	    if $good {
		lappend files_trim $i
	    }
	} else {
	    lappend files_trim $i
	}
    }
    
    # insert files in boxes
    set nbfiles [llength $files_trim]
    set nbrows [expr ceil($nbfiles /double($nbboxes))]
    set idx 0
    for { set i 0 } { $i < $nbboxes } { incr i 1 } {
	# fill the box
	$win.box$i delete 1.0 end
	for { set j 0 } { $j < $nbrows } { incr j } {
	    if { $idx < $nbfiles } {
		set f [lindex $files_trim $idx]
		incr idx
		if [file isdirectory $f] {
		    $win.box$i insert end $f directory
		} else {
		    $win.box$i insert end $f file 
		}
		$win.box$i insert end "\n"
	    }
	}
    }

}

################################################################
# MakeBrowser -
# Create a rudimentary browser with a certain number of columns.
# Each column is a (listbox-like) text widget with a single selection
# with the mouse.
# Used in conjunction with DisplayFiles.
# Bug: does not have scrollbars yet.
#
# Parameters:
# - nbcols	the number of columns for file/dir names
# - extns	a list of extensions that will appear in the "filter" menu
#
proc MakeBrowser { nbcols { extns {} } } {
    global BrowserAttributes

    toplevel .browser
    frame .browser.dir 
    label .browser.dir.label -text "Directory:" -width 10 -anchor w
    entry .browser.dir.name -width 60 -textvar BrowserAttributes(Dir)
    pack .browser.dir.label -side left -anchor w
    pack .browser.dir.name -side left -expand true -fill x
    pack .browser.dir -side top

    set f [frame .browser.ext]
    menubutton $f.label -text "Filter:" -width 10 -anchor w -padx 0 \
	-menu $f.label.menu -underline 0
    set m [menu $f.label.menu]
    foreach extn $extns {
	set laius [lindex $extn 0]
	set patterns [lindex $extn 1]
	$m add command -label "$laius ($patterns)" \
	    -command "set ext {$patterns}\nDisplayFiles .browser.list 2 . \$ext"
    }
    entry $f.name -width 60 -textvar BrowserAttributes(DefaultExt)
    pack $f.label -side left -anchor w
    pack $f.name -side left -expand true -fill x
    pack $f -side top

    frame .browser.list
    text .browser.list.box0 -width 30 -height 20
    text .browser.list.box1 -width 30 -height 20
    pack .browser.list.box0 .browser.list.box1 -side left \
	-expand true -fill both
    pack .browser.list -side top -expand true -fill both

    for { set i 0 } { $i < $nbcols } { incr i 1 } {
	# set bindings of the box, deactivating stantard Text bindings
	bindtags .browser.list.box$i {.browser all}

	.browser.list.box$i tag configure directory -font *-bold-*
	.browser.list.box$i tag configure directory -foreground green4
	.browser.list.box$i tag bind directory <1> \
	    "DisplayFiles .browser.list $nbcols \[ LineGet %W %x %y \] $BrowserAttributes(DefaultExt)"
	.browser.list.box$i tag bind file <Button-1> \
	    "LineSelect %W %x %y\nset BrowserAttributes(FileName) \[ LineGet %W %x %y \]"
	.browser.list.box$i tag bind file <Double-Button-1> \
	    "set BrowserAttributes(FileName) \[ LineGet %W %x %y \]\ndestroy .browser"
    }

    frame .browser.action
    button .browser.action.select -text "Select" -underline 0 \
	-command { destroy .browser }
    button .browser.action.cancel -text "Cancel" -underline 0 \
	-command "set BrowserAttributes(FileName) {}\n destroy .browser"
    pack .browser.action.select .browser.action.cancel -side left \
	-anchor center
    pack .browser.action -side top

    # bindings. Clear standard bindings in order to disable the
    # selection stuff.
    bind .browser <Escape> { .browser.action.cancel invoke }
    bind .browser.dir.name <Return> { DisplayFile .browser.list 2 $BrowserAttributes(Dir) $BrowserAttributes(DefaultExt) }
    bind .browser.ext.name <Return> { DisplayFile .browser.list 2 $BrowserAttributes(Dir) $BrowserAttributes(DefaultExt) }
    
    return .browser
}


proc LineGet { win xpos ypos } {

    set charpos [$win index @$xpos,$ypos]
    set line [$win get "$charpos linestart" "$charpos lineend"]

    return $line
}

proc LineSelect { win xpos ypos } {

    set charpos [$win index @$xpos,$ypos]
    $win tag remove sel 1.0 end
    $win tag add sel "$charpos linestart" "$charpos lineend"
    # puts "Selected in $win: [$win tag ranges sel]"
}

############################################################
# Procedure to specify a file name for reading.
# Uses the tk_getOpenFile if the Tcl version is recent enough.
#
proc GetOpenFile { options } {

    if { [info tclversion] > 7.5 } {
	# puts "Calling: {tk_getOpenFile} $options"
	set file_name [eval {tk_getOpenFile} $options]
    } else {
	set file_name [eval {BrowseFiles} [list $options]]
    }

    return $file_name
}

############################################################
# Procedure to specify a file name for saving.
# Uses the tk_getOpenFile if the Tcl version is recent enough.
#
proc GetSaveFile { options } {
    global save_name proposed_name

    if { [info tclversion] > 7.5 } {
	set save_name [eval {tk_getSaveFile} $options]
    } else {
	# set up the default value
	set save_name ""
	# try to guess an initial name for the file
	set proposed_name [GetOption {-initialfile} $options]

	set wn .savedialog
	if [winfo exists $wn] {
	    destroy $wn
	}
	if [catch {toplevel $wn}] { return }

	wm title $wn "Save"

	frame $wn.filename
	label $wn.filename.label -text "File Name" -width 15 -anchor w
	entry $wn.filename.field -textvar proposed_name -width 20 \
	    -relief sunken
	pack $wn.filename.label -side left 
	pack $wn.filename.field -side left -expand true -fill x

	frame $wn.actions
	button $wn.actions.browse -text Browse \
	    -command  "set proposed_name \[BrowseFiles \"$options\"\]" \
	    -underline 0
	button $wn.actions.save -text "Save" \
	    -underline 0 \
	    -command "if \[FileAccepted \$proposed_name\] { set save_name \$proposed_name\ndestroy $wn }"
	button $wn.actions.cancel -text "Cancel" \
	    -underline 0 \
	    -command "destroy $wn"
	pack $wn.actions.browse $wn.actions.save $wn.actions.cancel \
	    -side left

	bind $wn <Key-space> 	"$wn.actions.browse invoke"
	bind $wn <Key-question> "$wn.actions.browse invoke"
	bind $wn <Return> 	"$wn.actions.save   invoke"
	bind $wn <Escape> 	"$wn.actions.cancel invoke"

	pack $wn.filename -side top -fill x
	pack $wn.actions -side top

	tkwait window $wn

	unset proposed_name
	
    }

    return $save_name 
}

proc FileAccepted { name } {

    if [file exists $name] {
	switch [tk_dialog .warning "File Exists" "The file $name already exists. Do you want to overwrite it?" {warning} 0 {Overwrite} {Cancel}] {
	    0 { return 1 }
	    1 { return 0 }
	}
    } else {
	return 1
    }
}