#
# Graph creation module for gred
#
######################################################################
# Make_From_Geom -
# Creates a user dialog interface for setting up the parameters before
# the creation of a set of nodes and arcs with geometry.
#
# Parameters:
# - canvas		the canvas window in which to create the objects
# 
proc Make_From_Geom { canvas } {
    global GeomParams create_props

    set win .params

    if [catch {toplevel $win}] { return }

    wm title $win "Create Graph"

    # creating the geometry browser in $win.geom
    Geometry_Browser $win

    # a separator for the looks. Does not seem to work, though
    frame $win.sep -relief sunken -height 2

    # A zone for the weight of nodes
    frame $win.node
    label $win.node.label  -text "Node Weight" -width 20 -anchor w
    entry $win.node.weight -textvar create_props(weight) \
	-width 8 -relief sunken
    pack $win.node.label $win.node.weight -side left
    pack $win.node -side top -fill x

    # a zone for specifying the offset
    frame $win.offset
    label $win.offset.label -text "Offset (mm)" -width 17 -anchor w
    label $win.offset.xlabel -text "X=" -width 3 -anchor w
    entry $win.offset.xoffset -textvar GeomParams(xoffset) \
	-width 4 -relief sunken
    label $win.offset.ylabel -text "Y=" -width 3 -anchor w
    entry $win.offset.yoffset -textvar GeomParams(yoffset) \
	-width 4 -relief sunken
    pack $win.offset.label $win.offset.xlabel $win.offset.xoffset \
	$win.offset.ylabel $win.offset.yoffset -side left 

    # the zone for buttons
    frame $win.actions
    button $win.actions.ok -text "New" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n Create_From_Geom $canvas 0"
    button $win.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n Create_From_Geom $canvas 1"
    button $win.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n array set GeomParams {[array get GeomParams]}"
    pack $win.actions.ok $win.actions.merge $win.actions.cancel -side left

    # stacking the frames
    pack $win.geom    -side top -expand true -fill x
    pack $win.sep     -side top -expand true -fill x
    pack $win.offset  -side top -expand true -fill x
    pack $win.actions -side top -anchor center

    # standard bindings
    bind $win <Return> {$win.actions.ok invoke}
    bind $win <Escape> {$win.actions.cancel invoke}

}

######################################################################
# Create_From_Geom -
# The dipatching procedure for actual creation routines.
# It uses the values of GeomParams which have presumably been set
# with the geometry browser.
# Note that some parameters are passed on to the procedures, which
# use directly other parameters (offsets) from the global GeomParams!
# Returns a boolean value telling the success of the operation. Useful
# in Create_From_Matrix.
#
# Parameters:
# - win		the canvas window in which to create the objects
# - merge	a boolean telling whether the creation should be merged.
# 
proc Create_From_Geom { win merge } {
    global GeomParams

    if !$merge {
	if ![Clear_Graph $win] { return 0 }
    }

    switch $GeomParams(current) {
	cycle {
	    create_cycle $win $GeomParams(nodenb) $GeomParams(chordlist)
	}
	complete {
	    create_cycle $win $GeomParams(nodenb) \
		[iota 1 [expr $GeomParams(nodenb)/2]]
	}
	star {
	    create_star $win $GeomParams(nodenb)
	}
	random {
	    create_random $win $GeomParams(nodenb)
	}
	line {
	    create_string $win $GeomParams(nodenb) $GeomParams(orient)
	}
	grid {
	    create_grid $win $GeomParams(xdim) $GeomParams(ydim)
	}
	bipartite {
	    create_bipartite $win $GeomParams(leftnb) $GeomParams(rightnb) \
		$GeomParams(layout)
	}
	default {
	    ReportError "Geometry Error" "Geometry should not be \"$GeomParams(current)\".\nCannot create"
	}
    }

    return 1
}

proc Control_Panel { win } {
    global create_props GeomParams

    frame $win.node
    label $win.node.label  -text "Node Weight" -width 20 -anchor w
    entry $win.node.weight -textvar create_props(weight) \
        -width 8 -relief sunken
    pack $win.node.label $win.node.weight -side left
    pack $win.node -side top -fill x

    frame $win.offset
    label $win.offset.label -text "Offset (mm)" -width 17 -anchor w
    label $win.offset.xlabel -text "X=" -width 3 -anchor w
    entry $win.offset.xoffset -textvar GeomParams(xoffset) \
        -width 4 -relief sunken
    label $win.offset.ylabel -text "Y=" -width 3 -anchor w
    entry $win.offset.yoffset -textvar GeomParams(yoffset) \
        -width 4 -relief sunken
    pack $win.offset.label $win.offset.xlabel $win.offset.xoffset \
        $win.offset.ylabel $win.offset.yoffset -side left 
    pack $win.offset -side top -fill x

    return [list $win.node $win.offset]
}

######################################################################
# - create_circle -
# - create_cycle -
# - create_star -
# - create_random -
# - create_string -
# - create_grid -
# - create_bipartite -
# Procedures for actually creating graphs arranged according to some geometry.
# They (weakly) implement the principle of separation between the placement
# of nodes and that of arcs. Accordingly, all circle-based geometries
# call "create_circle", which fills the global "node_id" array with the ids 
# of the nodes created, indexed from 0 to the total number minus one.
# Coordinates will be recovered arterwards for creating the arcs.
#
# Parameters
# - win		the canvas where the nodes are created
# - others	easy to see
#
proc create_circle { win size } {
    global GeomParams node_id PI 
    global DEBUG

    if { $size <= 0 } { return }

    set Alpha [expr 2*$PI / $size]
    if [ expr { sin($Alpha/2.0) == 0.0 } ] {
	set Radius 0.0
    } else {
	set Radius [expr $GeomParams(object_distance) / sin( $Alpha/2 )]
    }

    # putting the center... at the center
    set val $GeomParams(xoffset)
    set xcenter [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)+$Radius]
    set val $GeomParams(yoffset)
    set ycenter [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)+$Radius]

    # perhaps it is necessary to resize the canvas?
    Resize_Graph $win [expr $xcenter+$Radius+$GeomParams(object_distance)] \
	[expr $ycenter+$Radius+$GeomParams(object_distance)]

    if $DEBUG {
	puts "Creating a circle with $size nodes."
    }

    # creation of the nodes 
    for { set i 0 } { $i < $size } { incr i 1 } {
	set xpos [xcoord $xcenter $Radius [expr $i*$Alpha]]
	set ypos [ycoord $ycenter $Radius [expr $i*$Alpha]]

	set node_id($i) [make_new_node $win $xpos $ypos]

    }
}

proc create_cycle { win size chordlist } {
    global GeomParams node_id
    global DEBUG

    create_circle $win $size

    if $DEBUG {
	puts "Creating a cycle with $size nodes and chords $chordlist..."
    }

    # creation of the chords
    foreach chord_length [Trim_Chordlist $size $chordlist] {

	if $DEBUG {
	    puts "doing chord $chord_length"
	}

	for { set i 0 } { $i < $size } { incr i 1 } {
	    
	    set k [expr ($i+$chord_length) % $size]

	    make_new_arc $win $node_id($i) $node_id($k)

	}
    }
    
    if $DEBUG {
	puts " ...Done"
    }
}

proc create_star { win nb_branches } {
    global GeomParams node_id PI
    global DEBUG

    if { $nb_branches < 0 } { return }

    if $DEBUG {
	puts "Creating of a star with $nb_branches branches."
    }

    create_circle $win $nb_branches

    # the creation of the circle has been handled by create_circle
    # but we have to determine where the center will be!
    set Alpha [expr 2*$PI / $nb_branches]
    if [ expr { sin($Alpha/2.0) == 0.0 } ] {
	set Radius 0.0
    } else {
	set Radius [expr $GeomParams(object_distance) / sin( $Alpha/2 )]
    }
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m] \
		   +$GeomParams(object_distance)+$Radius]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m] \
		   +$GeomParams(object_distance)+$Radius]

    # make center
    set node_id($nb_branches) [make_new_node $win $xorig $yorig]

    # put branches
    for { set i 0 } { $i < $nb_branches } { incr i 1 } {

	make_new_arc $win $node_id($nb_branches) $node_id($i)

    }
    if $DEBUG {
	puts " ...Done"
    }
}

proc create_random { win size } {
    global GeomParams node_id PI
    global DEBUG

    randomInit 1

    create_circle $win $size

    if $DEBUG {
	puts -nonewline "Creating a random graph with $size nodes"
    }

    for { set i 0 } { $i < $size } { incr i 1 } {
	for { set j [expr $i+1] } { $j < $size } { incr j 1 } {
	    if { [random] < $GeomParams(proba) } {

		make_new_arc $win $node_id($i) $node_id($j)

	    }
	}
    }
    if $DEBUG {
	puts " ...Done"
    }
}

proc create_string { win size orient } {
    global PI object_distance GeomParams 
    global DEBUG

    # defining the origin of the bounding rectangle
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m]+$object_distance]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m]+$object_distance]

    # a first loop for computing the position of each node
    for { set i 0 } { $i < $size } { incr i 1 } {

        case $orient {
            {hor} {
                set X($i) [expr $xorig + $object_distance*$i]
                set Y($i) $yorig
	    }
            {ver} {
		set X($i) $xorig
		set Y($i) [expr $yorig + $object_distance*$i]
	    }
            {dia} {
		set X($i) [expr $xorig + $object_distance*$i]
		set Y($i) [expr $yorig + $object_distance*$i]
	    }
	    {cir} {
		# I know: this should not be recomputed each time.
		set Alpha [expr 2*$PI / $size]
		if [ expr { sin($Alpha/2.0) == 0.0 } ] {
		    set Radius 0.0
		} else {
		    set Radius [expr $object_distance / 2 / sin( $Alpha/2 )]
		}
		set xorig [expr int(1.2*$Radius)]
		set yorig [expr int(1.2*$Radius)]
		
		set X($i) [xcoord $xorig $Radius [expr $i*$Alpha]]
		set Y($i) [ycoord $yorig $Radius [expr $i*$Alpha]]
	    }
	    {rec} {
		set xnb [expr int(sqrt($size))]
		set row [expr $i / $xnb]
		set col [expr $i % $xnb]
		set X($i) [expr $xorig + $object_distance*$row]
		set Y($i) [expr $yorig + $object_distance*$col]
	    }
	}
    }
    # find out if resizing is necessary
    set xmax [TableMax X 0 [expr $size-1]]
    set ymax [TableMax Y 0 [expr $size-1]]
    Resize_Graph $win \
	[expr $xmax+$object_distance] [expr $ymax+$object_distance]

    # now put the objects...
    for { set i 0 } { $i < $size } { incr i } {

	set xpos1 $X($i)
	set ypos1 $Y($i)

	set node_id($i) [make_new_node $win $xpos1 $ypos1]

    }

    # draw links
    for { set i 0 } { $i < [expr $size-1] } { incr i } {

	make_new_arc $win $node_id($i) $node_id([expr $i+1])

    }

    if $DEBUG {
	puts " ...Done"
    }
}

proc create_grid { win xdim ydim } {
    global GeomParams node_id
    global DEBUG

    if $DEBUG {
	puts "Creating a $xdim x $ydim grid."
    }

    set object_distance $GeomParams(object_distance)

    # putting the origin in the NW corner
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)]

    # perhaps it is necessary to resize the canvas?
    Resize_Graph $win [expr $xorig+$xdim*$GeomParams(xdistance)] \
	[expr $yorig+$ydim*$GeomParams(ydistance)]

    # put nodes on the grid, row by row
    set num 0
    for { set i 0 } { $i < $xdim } { incr i 1 } {
	for { set j 0 } { $j < $ydim } { incr j 1 } {
	    set xpos($num) [expr $xorig+$i*$GeomParams(xdistance)]
	    set ypos($num) [expr $yorig+$j*$GeomParams(ydistance)]

	    set node_id($num) [make_new_node $win $xpos($num) $ypos($num)]
	    incr num
	}
    }

    # put links, row by row also
    set num 0
    for { set i 0 } { $i < $xdim } { incr i 1 } {
	for { set j 0 } { $j < $ydim } { incr j 1 } {
	    if { $j < [expr $ydim-1] } {
		make_new_arc $win $node_id($num) $node_id([expr $num+1])
	    }
	    if { $i < [expr $xdim-1] } {
		make_new_arc $win $node_id($num) $node_id([expr $num+$ydim])
	    }
	    incr num
	}
    }

}

proc create_bipartite { win leftnb rightnb orient } {
    global object_distance GeomParams node_id
    global DEBUG

	set size [expr $leftnb + $rightnb]

    if { $orient == "cir" } {

	create_circle $win $size

    } else {
	# defining the origin of the bounding rectangle
	set val $GeomParams(xoffset)
	set xorig [expr [winfo pixels . ${val}m]+$object_distance]
	set val $GeomParams(yoffset)
	set yorig [expr [winfo pixels . ${val}m]+$object_distance]

	# computing the dimensions of the rectangle. The golden ratio
	# is used for the width.
	if { $rightnb >= $leftnb } {
	    set maxnb $rightnb
	    set leftshift [expr ($maxnb-$leftnb)/2.0]
	    set rightshift 0
	} else {
	    set maxnb $leftnb
	    set rightshift [expr ($maxnb-$rightnb)/2.0]
	    set leftshift 0
	}	    
	set height [expr $maxnb * $object_distance]
	set width [expr $height*0.6]

	# a first loop for computing the position of each node
	for { set i 0 } { $i < $leftnb } { incr i 1 } {
	    set ii [expr $i + $leftshift]
	    case $orient {
		{hor} {
		    set X($i) [expr $xorig + $object_distance*$ii]
		    set Y($i) $yorig
		}
		{ver} {
		    set X($i) $xorig
		    set Y($i) [expr $yorig + $object_distance*$ii]
		}
	    }
	}
	for { } { $i < $size } { incr i 1 } {
	    set ii [expr $i-$leftnb+$rightshift]
	    case $orient {
		{hor} {
		    set X($i) [expr $xorig + $object_distance*$ii]
		    set Y($i) [expr $yorig + $width]
		}
		{ver} {
		    set X($i) [expr $xorig + $width]
		    set Y($i) [expr $yorig + $object_distance*$ii]
		}
	    }
	}

	# find out if resizing is necessary
	set xmax [TableMax X 0 [expr $size-1]]
	set ymax [TableMax Y 0 [expr $size-1]]
	Resize_Graph $win \
	    [expr $xmax+$object_distance] [expr $ymax+$object_distance]

	# now put the objects...
	for { set i 0 } { $i < $size } { incr i } {

	    set xpos1 $X($i)
	    set ypos1 $Y($i)

	    set node_id($i) [make_new_node $win $xpos1 $ypos1]

	}
    }

    # draw links
    for { set i 0 } { $i < $leftnb } { incr i } {
	for { set j 0 } { $j < $rightnb } { incr j } {

	    make_new_arc $win $node_id($i) $node_id([expr $j+$leftnb])

	}
    }

    if $DEBUG {
	puts " ...Done"
    }
}

proc make_cycle { win } {
    global size

    if [catch {toplevel .params}] { return }

    wm title .params "Cycle Parameters"

    frame .params.size
    label .params.size.label -text "Number of Nodes:" -width 15  -anchor w
    entry .params.size.field -textvar size -width 3 -relief sunken
    frame .params.chord
    label .params.chord.label -text "List of chords:" -width 15 -anchor w
    entry .params.chord.field -textvar chordlist -width 10 -relief sunken

    pack .params.size.label .params.size.field -side left -fill x
    pack .params.chord.label .params.chord.field -side left -fill x

    Control_Panel .params

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\create_cycle $win \$size \$chordlist"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.size .params.chord .params.actions -side top

}

proc make_complete { win } {

    global size

    if [catch {toplevel .params}] { return }

    frame .params.size
    label .params.size.label -text "Number of Nodes:"
    entry .params.size.field -textvar size -width 3 -relief sunken

    pack .params.size.label .params.size.field -side left

    Control_Panel .params

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\create_cycle $win \$size \[iota 1 \[expr \$size/2\]\]"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.size .params.actions -side top

}

proc make_star { win } {

    global size

    if [catch {toplevel .params}] { return }

    frame .params.size
    label .params.size.label -text "Number of branches:"
    entry .params.size.field -textvar size -width 3 -relief sunken

    pack .params.size.label .params.size.field -side left

    Control_Panel .params

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\create_star $win \$size"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.size .params.actions -side top

}

proc make_random { win } {
    global size GeomParams

    if [catch {toplevel .params}] { return }

    frame .params.size
    label .params.size.label -text "Number of nodes:" -width 20 -anchor w
    entry .params.size.field -textvar size -width 4 -relief sunken
    frame .params.proba
    label .params.proba.label -text "Edge probability:" -width 20 -anchor w
    entry .params.proba.field -textvar GeomParams(proba) \
	-width 12 -relief sunken

    pack .params.size.label .params.size.field -side left
    pack .params.proba.label .params.proba.field -side left
    pack .params.size .params.proba -side top -fill x

    Control_Panel .params

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\create_random $win \$size"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.actions -side top

}

proc make_string { win } {

    if [catch {toplevel .params}] { return }

    frame .params.size
    label .params.size.label -text "Number of nodes:" -width 20 -anchor w
    entry .params.size.field -textvar size -width 3 -relief sunken

    frame .params.orient
    label .params.orient.label -text "Orientation:" -width 20  -anchor w
    radiobutton .params.orient.hor -text "Horizontal"  -variable orient \
	-value hor
    radiobutton .params.orient.ver -text "Vertical"    -variable orient \
	-value ver
    radiobutton .params.orient.dia -text "Diagonal"    -variable orient \
	-value dia
    radiobutton .params.orient.cir -text "Circular"    -variable orient \
	-value cir
    radiobutton .params.orient.rec -text "Rectangular" -variable orient \
	-value rec
    pack .params.orient.label .params.orient.hor .params.orient.ver \
	.params.orient.dia .params.orient.cir .params.orient.rec \
	-side left -fill x

    pack .params.size.label .params.size.field -side left

    Control_Panel .params

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\create_string $win \$size \$orient"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.size .params.orient .params.actions -side top

}

proc Create_From_Matrix { win file_name } {
    global GeomParams weight node_id
    global DEBUG

    if ![file exists $file_name] {
	ReportError "Create Error" "Can't open file $file_name."
	return
    }
    # get matrix
    set matfile [open $file_name r]
    set matrix [read $matfile]
    if { [lindex $matrix 0] == "sp" } {
	set size [lindex $matrix 1]
	set SPARSE 1
	set matrix [lrange $matrix 2 end]
    } else {
	set size [lindex $matrix 0]
	set SPARSE 0
	set matrix [lrange $matrix 1 end]
    }

    if { $size < 1 } { 
	ReportError "Create Error" "Impossible size: $size."
	return
    }

    if { $DEBUG } {
	puts -nonewline "creation of an interaction graph from matrix file $file_name"
    }

    # creating the nodes on a circle
    create_circle $win $size

    # extract and set weights
    for { set i 0 } { $i < $size } { incr i 1 } {
	set weight($node_id($i)) [lindex $matrix $i]
    }
    set matrix [lrange $matrix $size end]

    # extracting arcs from the matrix 
    set arcs {}
    if $SPARSE {
	set nbarcs [expr [llength $matrix] / 2]
	if [expr 2*$nbarcs != [llength $matrix]] {
	    ReportWarning "Create from Matrix" \
		"Unexpected odd number of element\nin sparse matrix.\nTrying anyway..."
	}
	# reading arcs discarding self loops
	for { set i 0 } { $i < $nbarcs } { incr i 1 } {
	    set from [lindex $matrix [expr 2*$i]]
	    set to [lindex $matrix [expr 2*$i+1]]
	    if [expr $from != $to ] {
		lappend arcs [list $from $to]
	    }
	}
    } else {
	if [expr $size*$size < [llength $matrix]] {
	    ReportWarning "Create from Matrix" \
		"Unexpected number of element\nin full matrix.\nTrying anyway..."
	} elseif [expr $size*$size > [llength $matrix]] {
	    ReportError "Create Error" \
		"Not enough elements\nin full matrix.\nAbort..."
	    for  { set i 0 } { $i < $size } { incr i 1 } {
		Delete_Object $win $node_id($i)
	    }
	    return
	}
	# extract arcs according to the matrix, taking only above the diagonal
	for { set i 0 } { $i < $size } { incr i 1 } {
	    for { set j [expr $i+1] } { $j < $size } { incr j 1 } {
		set val [lindex $matrix [expr $i*$size+$j]]
		if { $val == "1" } {
		    lappend arcs [list $i $j]
		}
	    }
	}
    }
	
    # draw arcs 
    foreach arc $arcs {
	set i [lindex $arc 0]
	set j [lindex $arc 1]

	make_new_arc $win $node_id($i) $node_id($j)

    }
    if $DEBUG {
	puts " ...done"
    }
}

######################################################################
# Make_From_Matrix -
# Creates a user dialog interface for choosing an adjacency matrix
# file and setting up the parameters before the creation of the representation.
# At the moment, no geometry browser is proposed but this will come!
#
# Parameters:
# - canvas	the canvas window in which to create the objects
# 
proc Make_From_Matrix { win } {

    set types {
	{{Adjacency Matrix}       {.adj .mat}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getOpenFile -defaultextension adj \
# 		       -filetypes $types -parent .]
    set options "-defaultextension adj -filetypes {$types} -parent ."

    set file_name [GetOpenFile $options]

    if { $file_name != "" } {
	Create_From_Matrix $win $file_name
    }

}
