#
# Module grouping the procedures for "quantitative" analysis of Petri Nets
#
# - deterministic (max,+) methods
# - Poisson-driven methods
# - simulation
#
######################################################################
# Max_Plus_Solve -
# Interface Procedure for deterministic (max,+) solution methods.
# Several properties of the current graph can be computed.
# Applies only to Event Graphs with deterministic firing/holding times.
#
# Parameters:
# - property	the (max,+) spectral properties to compute. One of: 
#		transients, cycle, graph.
#
proc Max_Plus_Solve { property } {
    global number2id
    global KEEP DEBUG

    set TMP_EGL_FILE "/tmp/pned[pid].egl"
    set TMP_RES_FILE "/tmp/pned[pid].res"
    set TMP_ARC_FILE "/tmp/pned[pid].arc"
    set TMP_MAT_FILE "/tmp/pned[pid].mat"

    set scscyc "scscyc"

    if ![Is_Max_Plus] {
	ReportError "Analysis Error" "Not an Event Graph or not Determinisic."
	return
    }

    write_matrix $TMP_MAT_FILE bruno "A" full yes 0

    # some effort is required to cope with decimal entries. I choose
    # to apply a multiplyer of 1000 by default.
    set command "$scscyc -f $TMP_MAT_FILE -o $TMP_RES_FILE -precision 1000"

    switch $property {
	"transients" {
	    LogResult "Executing $command -transient" ""
	    if [catch {eval exec "$command -transient"} problem] {
		LogResult "Analysis Error" "" "$problem" 
	    }
	}
	"cycle" -
	"graph" {
	    LogResult "Executing $command" ""
	    if [catch {eval exec "$command"} problem] {
		LogResult "Analysis Error" "" "$problem"
	    }
	}
	"critical" {
	    set command \
		"$scscyc -f $TMP_MAT_FILE -o $TMP_ARC_FILE -precision 1000"

	    LogResult "Executing $command -cycles" ""
	    if [catch {eval exec "$command -cycles"} problem] {
		LogResult "Analysis Error" "" "$problem"
	    } else {
		# remake EGL file...
		write_EGL $TMP_EGL_FILE 1

		set command \
		    "eg_crit_path -f $TMP_EGL_FILE -arcs $TMP_ARC_FILE -o $TMP_RES_FILE"
		LogResult "Executing $command" ""
		if [catch {eval exec "$command"} problem] {
		    LogResult "Analysis Error" "" "$problem"
		}
	    }
	}
    }
    # Even if an error has been encountered, we have not exited yet.
    # This is because some errors of scscyc are just warnings, and
    # a result may have been produced anyway.
    # So proceed with displaying the result... if any.
    if [catch {set resfile [open $TMP_RES_FILE r]}] { return }

    # There is indeed. Displaying the result 
    set result [read $resfile]
    close $resfile 

    switch $property {
	"cycle" {
	    set match [regsub \
			   {.*cycle mean: ([0-9]*)[ ]*/[ ]*([0-9]*).*} \
			   $result {\1/\2} lyapu]
	    set match [regsub {.*cycle mean: ([0-9]*)[ ]*/[ ]*([0-9]*).*} \
			   $result {\1.0/\2.0} fract]
	    set result "Cycle time = $lyapu ([expr $fract])\n"
	}
	"transients" {
	    set match [regsub {[^T]*} $result {} lyapu]
	    set result "Transients:\n$lyapu"
	}
	"graph" {
	    set match [regsub {([^C]*)} $result {} lyapu]
	    set result $lyapu
	}
	"critical" {
	    set match [regsub -all -- "-." $result {} result]
	    set match [regsub -all "\[mv\]" $result {} nodes]
	    set places {}
	    set trans {}
	    set trids {}
	    set plids {}
	    set liids {}
	    for { set i 0 } { $i < [llength $nodes] } { incr i 5 } {
		set orig [lindex $nodes $i]
		set dest [lindex $nodes [expr $i+2]]
		set origid $number2id([expr $orig+1])
		set destid $number2id([expr $dest+1])
		lappend trids $origid
		lappend places [lindex $nodes [expr $i+1]]
		set plouts [Get_Trans_Outputs $origid]
		foreach place $plouts {
		    set trouts [Get_Place_Outputs $place]
		    if { -1 != [lsearch -exact $trouts $destid] } {
			lappend plids $place
		    }
		}
	    }

	    set trans [lsort $trans]
	    set places [lsort $places]

	    set ids [concat $trids $plids]
	    UnHighlight_Objects .graph.canvas [.graph.canvas find withtag HIGH]
	    Highlight_Objects .graph.canvas $ids

	    switch [tk_dialog .graph.property "Highlighting" "Reset original colors now?" {} 0 {Yes} {No}] {
		0 { UnHighlight_Objects .graph.canvas $ids }
		1 { return }
	    }

	    set result "Critical Trans: $trans ($trids)\nCritical Places: $places ($plids)"

	}
    }

    LogResult "Spectral Analysis Results" "$result"

    if !$KEEP {
	RemoveFile $TMP_EGL_FILE
	RemoveFile $TMP_RES_FILE
	RemoveFile $TMP_ARC_FILE
	RemoveFile $TMP_MAT_FILE
    }
}

######################################################################
# Simulate -
# Dialog interface to let the user specify parameters for the 
# simulation of the Petri Net (presently: only an Event graph).
#
# Parameters: none
#
proc Simulate { } {
    global sim_length gen_file_name

    if ![Is_Event_Graph] {
	ReportError "Analysis Error" "Not an Event Graph."
	return 
    } elseif { 0 != [string compare "1" "[Is_FIFO]"] } {
	ReportError "Analysis Error" "This Net may not be FIFO"
	return
    } else {
	if [catch {toplevel .params}] { return }

	wm title .params "Simulation Parameters"

	if [Contains_Sources] {
	    set with with
	    frame .params.sources
	    radiobutton .params.sources.with -text "With Sources" \
		-variable with_or_without -value with
	    radiobutton .params.sources.without -text "Without Sources" \
		-variable with_or_without -value without
	    pack .params.sources.with .params.sources.without -side left
	    pack .params.sources
	}

	frame .params.simlength
	label .params.simlength.label -text "Simulation Length" \
	    -width 15 -anchor w
	entry .params.simlength.field -textvar sim_length -width 8 \
	    -relief sunken
    
	frame .params.actions
	button .params.actions.ok -text "Ok" \
	    -command "destroy .params\n Run_Simulation \$sim_length \$with_or_without"
	button .params.actions.cancel -text "Cancel" \
	    -command "destroy .params"
    
	pack .params.simlength.label .params.simlength.field -side left \
	    -fill x
	pack .params.actions.ok .params.actions.cancel -side left
	pack .params.simlength -side top -fill x
	pack .params.actions -side top

	bind .params <Return> {.params.actions.ok invoke}
	bind .params <Escape> {.params.actions.cancel invoke}

	grab .params
    }
}

######################################################################
# Run_Simulation -
# Interface to the call of "eg_sim" which simulates the Event Graph.
# The application is called with its option -compile in order to
# speed up **long** simulations (disabled for the time being).
#
# Parameters: 
# - sim_length	the number firings to compute
# - with	a flag telling whether sources should be included (value
#		"with). They are by default. 
#
proc Run_Simulation { sim_length { with with } } {

    set TMP_EGL_FILE "/tmp/pned[pid].egl"
    set TMP_RES_FILE "/tmp/pned[pid].res"

    if ![write_EGL $TMP_EGL_FILE [expr {$with == "with"}]] {
	ReportError "Simulation Error" "Could not save network as EGL"
	return
    }

# "compile" removed while the problem is not fixed.
#     if [catch "exec eg_sim -f $TMP_EGL_FILE -compile -nevt $sim_length > $TMP_RES_FILE 2> /dev/null" problem] {
# 	LogResult "Analysis Error" "" "$problem"
# 	return
#     }
    if [catch "exec eg_sim -f $TMP_EGL_FILE -nevt $sim_length > $TMP_RES_FILE 2> /dev/null" problem] {
	LogResult "Analysis Error" "" "$problem"
	return
    }

    set resfile [open $TMP_RES_FILE r]
    set result [read $resfile]
    close $resfile

    LogResult "Simulation Results" $result
}

######################################################################
# Solve_Poisson -
# Procedure driving the solution of Poisson-Deterministic systems.
# The solution consists in calling m_eg and displaying the resulting
# curves with gnuplot. Before doing this, it is necessary to perform
# the (max,+) analysis (with transient) on the deterministic part of
# the network.
#
# Bug: in order to make the gnuplot window go away, it seems necessary
# to issue a carrier-return to pned in the calling window.
#
# Parameters: none
#
proc Solve_Poisson { } {
    global source_list source_rate trans_list
    global name env
    global KEEP

    set TMP_RES_FILE "/tmp/pned[pid].res"
    set TMP_MAT_FILE "/tmp/pned[pid].mat"
    set TMP_DAT_FILE "/tmp/pned[pid].mk1"
    set TMP_GNU_FILE "/tmp/pned[pid].gnu"
            
    set scscyc "scscyc"
    if [info exists env(GNUPLOT)] {
	set gnuplot $env(GNUPLOT)
    } else {
	set gnuplot "gnuplot"
    }

    if ![Is_Event_Graph] {
	ReportError "Analysis Error" "Not an Event Graph."
	return 
    }

    set flaglist [.graph.canvas find withtag selected]
    set sellist {}
    foreach node $flaglist {
	if [Is_Type $node TRANS] {
	    set sellist [concat $sellist $node]
	}
    }
    puts "Selected trans=$sellist"

    if { [llength $sellist] == 0 } {
	ReportError "Analysis Error" "No selected transitions..."
	return
    }

    # Will need to recover the number of transitions in the EGL!
    set curtrans 1
    foreach trans $trans_list {
	set number([lindex $trans 0]) $curtrans
	incr curtrans 1
    }

    # determine entry point of the source. Don't check if
    # it is unique. To be done later.
    set source_id [lindex [lindex $source_list 0] 0]
    set the_place [lindex [Get_Trans_Outputs $source_id] 0]
    set the_input [lindex [Get_Place_Outputs $the_place] 0]
    set flagged_trans $number($the_input)

    # checking for finiteness of arrival rate
    if { $source_rate($source_id) <= 0 } {
	ReportError "Analysis Error" "Zero interarrivals at source\nCannot analyze."
	return 
    }


    # preparing gnuplot instructions. There could be a control pannel here.
    # so set some options up...
    set gnufile [open $TMP_GNU_FILE w]
    puts $gnufile "plot \[\]\[0:1\] \\"

    # computing the matrix and the transients. Depends only on the
    # input transition, not on the measured ones.
    write_matrix $TMP_MAT_FILE bruno "A" full yes $flagged_trans
    exec $scscyc -f $TMP_MAT_FILE -o $TMP_RES_FILE -transient
    
    set resfile [open $TMP_RES_FILE r]
    set result [read $resfile]
    close $resfile
    # need also the size of the flattened matrix
    set resfile [open $TMP_MAT_FILE r]
    set size [gets $resfile]
    puts "sz=$size"
    close $resfile

    # extract all interesting parameters...
    regsub {.*cycle mean: ([0-9]*)[ ]*/[ ]*([0-9]*).*} \
	$result {\1.0/\2.0} lyapu
    puts "li=$lyapu"
    # actually, don't need it! File needs lambda!

    regsub {.*period:[ ]*([0-9]*).*} $result {\1} translg
    puts "tr=$translg"

    regsub {.*Cyclicity: cyc = ([0-9]*).*} $result {\1} cycl
    puts "cy=$cycl"

    regsub {[^T]*(T = .*)length.*} $result {\1} trans
    puts "tr=$trans"
    
    # method to get back the numbers: read all the transient analysis,
    # get rid of non-numeric entries (except eps!) and put them in 
    # a list. Other "line-oriented" approaches have failed...
    regsub -all {T = [^:]*:} $trans {} numbers
    regsub -all {[^\.0-9eps]+} $numbers { } numbers
    set numbers [string trim $numbers { }]
    set numberl [split $numbers]
    puts "Values=$numberl"

    if { [llength $numberl] != [expr ($translg+$cycl)*$size] } {
	ReportError "Analysis Error" "Bad number of numbers..."
	return }

    set rg 1
    foreach tr_id $sellist {

	set datafile [open "$TMP_DAT_FILE.$rg" w]
	puts $datafile [expr 1.0/$source_rate($source_id)]
	puts $datafile [expr $translg-1]
	puts $datafile $cycl

	for { set i 0 } { $i < $translg } { incr i 1 } {
	    set idx [expr $number($tr_id)+$i*$size-1]
	    puts $datafile [lindex $numberl $idx]
	}
	set idx [expr $number($tr_id)-1+($translg-1)*$size]
	set offset [lindex $numberl $idx]
	for { set i $translg } { $i < [expr $translg+$cycl] } { incr i 1 } {
	    set idx [expr $number($tr_id)+$i*$size-1]
	    puts $datafile [expr [lindex $numberl $idx]-$offset]
	    set offset [lindex $numberl $idx]
	}
    
	# stationary distrib required
	puts $datafile -1
	close $datafile

	exec mk1 -f "$TMP_DAT_FILE.$rg" -o "$TMP_RES_FILE.$rg"
	
	puts -nonewline $gnufile "\"$TMP_RES_FILE.$rg\""
	puts -nonewline $gnufile " title \"Trans $tr_id"
	if [info exists name($tr_id)] {
	    if { $name($tr_id) != "" } {
		puts -nonewline $gnufile " ($name($tr_id))"
	    }
	}
	puts -nonewline $gnufile "\"  with lines"
	if { $rg < [llength $sellist] } {
	    puts $gnufile ",\\"
	} else {
	    puts $gnufile ""
	}
	incr rg

    }

    puts $gnufile "pause -1"
    close $gnufile

    # done with the files. plotting

    exec $gnuplot $TMP_GNU_FILE

    # finished. Ouf. Some cleanup
    if !$KEEP {
	RemoveFile $TMP_MAT_FILE
	for { set rg 1 } { $rg <= [llength $sellist] } { incr rg 1 } {
	    RemoveFile "$TMP_DAT_FILE.$rg"
	    RemoveFile "$TMP_RES_FILE.$rg"
	}
	RemoveFile $TMP_GNU_FILE
    }

}
