#define CST(x)		x
#define EXP(x)		Exponential(x)
#define BER(x)		Bernoulli(x)
#define UNI(x,y)	Uniform(x,y)
#define ERL(x,y)	Erlang(x,y)
#define HYP(x)		HyperExpo(x)

#define MAX_RANDOM_VALUE          0x7FFFFFFF

/****************************************************************/
/* MIN and MAX functions not as macros, so that they can be	*/
/* used with functions drawing random numbers.			*/
/****************************************************************/
double MAX(double x, double y)
{
  return( x < y ? y : x );
}

double MIN(double x, double y)
{
  return( x > y ? y : x );
}

/****************************************************************/
/* Drawing of random numbers according to several classical	*/
/* distributions.						*/
/****************************************************************/
double U_0_1(void)
{
  return (double)random()/MAX_RANDOM_VALUE;
}

double Uniform(double a, double b)
{
  return a + (b-a)*U_0_1();
}

double Exponential(double Mean)
{
  return -Mean*log(U_0_1());
}

double Bernoulli(double Proba)
{
  if (U_0_1() < Proba)
    return 1.0;
  else
    return 0.0;
}

double Geometric(double Proba)
{
  if ( Proba == 0.0 )
    return 0.0;
  else if ( Proba < 1 )
    return floor( Exponential( -1/log(Proba)) );
  else {
    fprintf(stderr, "Error: Geometric with proba=1. 0.0 assumed.\n");
    return 0.0;
  }
}

double Erlang(double N, double Mean)
/* N is kept double to avoid conversions */
{
  register int		i;
  register double	U;

  U = U_0_1();
	
  for (i=1; i<N; i++)	/* even if N not integer, this should	*/
			/* be the correct number of iterations	*/
    U *= U_0_1();

  return -Mean*log(U)/N;
}

double HyperExpo(double *Parameters)
{
  register double	U;
  register int    	i;

  U = U_0_1();

  for (i=0; (i<Parameters[0]) && (U>Parameters[i+1]); i+=2 )
    U -=Parameters[i+1];

  return Exponential( Parameters[i+2] );
}

double Discrete(double *Parameters)
{
  register double	U;
  register int    	i;
  register int		NbVals;

  NbVals = (int) Parameters[0];

  U = U_0_1();

  for ( i=NbVals+1; (i<=2*NbVals) && (U>Parameters[i]); i++ )
    U -=Parameters[i];

  return Parameters[i-NbVals];
}

/* Note: the generation of Weibull sambles wouls be faster if log(a)	*/
/* and 1/nu were stored instead of a and nu.				*/
double	Weibull(double a, double nu)
{
  register double	U;

  U = U_0_1();
  
  return pow(-log(U)/log(a),1.0/nu);
}

/* Generation of an exponential using the Polar Method (Knuth v2 p. 117) */
double Normal(double Mean, double StdDev )
{
  register double	v1;
  register double	v2;
  register double	S;

  /* Generate uniform on the unit circle */
  do { 
    v1 = 2 * U_0_1() - 1;
    v2 = 2 * U_0_1() - 1;
   } while ( 1 < ( S = v1*v1+v2*v2 ) );

  return Mean + StdDev * v1 * sqrt( -2.0 * log(S)/S );
}

/* Lognormal = exp(Normal). Adjusting the mean and var requires 	*/
/* computing parameters of the Normal. Would be faster if pre-computed.	*/
double LogNormal(double Mean, double StdDev )
{
  register double	m;
  register double	s2;

  s2 = log( 1.0 + (StdDev / Mean)*(StdDev / Mean) );
  m = log( Mean ) - s2 / 2;

  return exp( Normal( m, sqrt(s2) ) );
}

double Pareto( double a, double Alpha )
{
  return a * ( pow( U_0_1(), -1.0/Alpha ) - 1.0 );
}

/****************************************************************/
/* Procedure:		draw_iid_table				*/
/* Fill a table with random values drawn from a given law.	*/
/* Parameters:							*/
/*	The_Val:	array of double, returned filled	*/
/*	The_Size:	its size				*/
/*	The_Law:	the probability law according to which	*/
/*			the random numbers are drawn		*/
/****************************************************************/
void draw_iid_table(double	*The_Val,
		    int		The_Size,
		    LAW_DESC	The_Law )
{
  int	i;

  switch( The_Law.Name ) {
  case 'B':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Bernoulli( The_Law.Parameters[0] );
    break;
  case 'C':
    for (i=0; i<The_Size; i++)
      The_Val[i] = The_Law.Parameters[0];
    break;
  case 'D':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Discrete( The_Law.Parameters );
    break;
  case 'E':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Exponential( The_Law.Parameters[0] );
    break;
  case 'g':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Geometric( The_Law.Parameters[0] );
    break;
  case 'G':
    for (i=0; i<The_Size; i++)
      The_Val[i] = 1.0 + Geometric( The_Law.Parameters[0] );
    break;
  case 'H':
    for (i=0; i<The_Size; i++)
      The_Val[i] = HyperExpo( The_Law.Parameters );
    break;
  case 'L':
    for (i=0; i<The_Size; i++)
      The_Val[i] = LogNormal( The_Law.Parameters[0],
			      The_Law.Parameters[1] );
    break;
  case 'N':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Normal( The_Law.Parameters[0],
			   The_Law.Parameters[1] );
    break;
  case 'p':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Pareto( The_Law.Parameters[0],
			   The_Law.Parameters[1] );
    break;
  case 'R':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Erlang( The_Law.Parameters[0],
			   The_Law.Parameters[1] );
    break;
  case 'U':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Uniform( The_Law.Parameters[0],
			    The_Law.Parameters[1] );
    break;
  case 'W':
    for (i=0; i<The_Size; i++)
      The_Val[i] = Weibull( The_Law.Parameters[0],
			    The_Law.Parameters[1] );
  default:
    printf("Warning: unsupported law: %c. 0 values assumed.\n",
	   The_Law.Name );
    for (i=0; i<The_Size; i++)
      The_Val[i] = 0.0;		
  }
}

/****************************************************************/
/* Procedure 		init_simulation				*/
/*	Makes initializations, draws initial data tables	*/
/* Parameters:							*/
/*	Reseau		the net to be simulated			*/
/****************************************************************/
void init_simulation(void)
{
  int		i;

  /* Initialize date tables */
  for (i=0; i<NB_TRANS; i++)
    if ( Tr_Draw_It[i] )
      draw_iid_table( S[i], BUF_SIZE, Tr_Law[i] );

  for (i=0; i<NB_PLACES; i++)
    if ( Pl_Draw_It[i] )
      draw_iid_table( L[i], BUF_SIZE, Pl_Law[i] );

}

/****************************************************************/
/* Procedure:		wrap_buffers				*/
/* handles the circular buffer when it is full: store stats,	*/
/* dump trace, draws new batch of random numbers.		*/
/* Parameters:							*/
/*	None.							*/
/****************************************************************/
void	wrap_buffers(void)
{
#ifdef Trace
  int		Num;
  int		j;
  boolean 	Firing_Ok;

  /* looks for the trace */
  More_Events = TRUE;
  Was_X_Evt = TRUE;
  Was_Y_Evt = TRUE;

  New_Date = MIN( Min_X_Val, Min_Y_Val );
  Old_Date = New_Date;

  while ( More_Events ) {

    if ( Was_X_Evt ) {	/* look for next X event */
      Arg_Min_X = 0;
      Min_X_Val = X[0][Cur_X_Evt[0]];
      for ( Num=1 ; Num < NB_TRANS ; Num++ )
	if ( X[Num][Cur_X_Evt[Num]] < Min_X_Val ) {
	  Arg_Min_X = Num;
	  Min_X_Val = X[Num][Cur_X_Evt[Num]];
	}
    }
    if ( Was_Y_Evt ) {	/* look for next Y event, taking taboos */
			/* into account.			*/
      for ( Num=0; Taboo_Trans[Num]; Num++ );
      Arg_Min_Y = Num;
      Min_Y_Val = Y[Num][Cur_Y_Evt[Num]];
      for ( ; Num < NB_TRANS ; Num++ )
	if ( !Taboo_Trans[Num] 
	     && ( Y[Num][Cur_Y_Evt[Num]] < Min_Y_Val ) ) {
	  Arg_Min_Y = Num;
	  Min_Y_Val = Y[Num][Cur_Y_Evt[Num]];
	}
    }

    New_Date = MIN( Min_X_Val, Min_Y_Val );	
    if ( New_Date > Old_Date )
      Old_Date = New_Date;

    if ( ( Min_X_Val < Min_Y_Val ) && ( Min_X_Val < Trace_End ) ) {

      if ( Debug )
	printf("t=%f Evt X en %s\n", Min_X_Val, 
	       gss_trans_name[ Arg_Min_X ] );
      CODE1("MODIFY %s \"status\" = 1\n", gss_trans_name[ Arg_Min_X ] );
      CODE1("MODIFY TheClock \"ctime\" = %7.3f\n", Min_X_Val );
      CODE("eol+\n");
      Cur_X_Evt[ Arg_Min_X ] ++;
      More_Events = ( Cur_X_Evt[ Arg_Min_X ] != TMAX );
      Was_X_Evt = TRUE;
    }
    else if ( Min_Y_Val < Trace_End ) {

      if ( Debug )
	printf("t=%f Evt Y en %s\n", Min_Y_Val, gss_trans_name[ Arg_Min_Y ] );
      CODE1("MODIFY TheClock \"ctime\" = %7.3f\n", Min_Y_Val );

      Firing_Ok = TRUE;

#include "egc_end_firing_trace.c"

      CODE("eol+\n");

      if ( Firing_Ok ) {
	Cur_Y_Evt[ Arg_Min_Y ] ++;
	More_Events = ( Cur_Y_Evt[ Arg_Min_Y ] != TMAX );
	Was_Y_Evt = TRUE;
	/* Resetting the taboo flags */
	if ( Some_Taboo )
	  for ( j=0; j<NB_TRANS; j++ )
	    Taboo_Trans[j] = FALSE;
      }
      else {
	/* Transition becomes taboo to avoid looking at it again now */
	Taboo_Trans[ Arg_Min_Y ] = TRUE;
	Some_Taboo = TRUE;
	if ( Debug ) {
	  printf("Unfeasible evt detected for trans %d at t=%f\n", 
		 Arg_Min_Y, Min_Y_Val );
	}
      }
    }
    else   
      More_Events = FALSE;
  }
#endif

  init_simulation();
}
